#!/usr/bin/perl

#    figbubbles.pl - Figmap drawing utility
#    Copyright (C) 1999 Michael Jung
#    e-mail: miju@phantasia.org
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# Creates circles on fractal lines of fig polylines. Works on all
# closed and open polylines of an input file (stdin/stdout if
# unspecified) of a specified color and depth or all, if
# unspecified. The used polylines are removed. Leaves the rest of the
# fig file untouched. Useful with forest boundaries. Be sure to have
# the right orientation of the polyline, since indentation to the
# right and left of the line are not equal by default.

### ARGS SECTION

$LVL = 4;
$IN = 0.1443;
$OUT = 0.0722;

$Infile = "-";
$Outfile = "-";
$Color = -2;
$Depth = -1;

while ($#ARGV >= 0) {
    $arg = shift(@ARGV);
    if ($arg eq "-i") { $Infile = shift(@ARGV); }
    elsif ($arg eq "-o") { $Outfile = shift(@ARGV); }
    elsif ($arg eq "-d") { $Depth = shift(@ARGV); }
    elsif ($arg eq "-c") { $Color = shift(@ARGV); }
    elsif ($arg eq "-right") { $IN = shift(@ARGV); }
    elsif ($arg eq "-left") { $OUT = shift(@ARGV); }
    elsif ($arg eq "-level") { $LVL = shift(@ARGV); }
    elsif ($arg eq "-seed") { srand(shift(@ARGV)); }
    else {
        print "Usage: figbubbles.pl [-help] [-i infile] ";
        print "[-o outfile] [-d depth] [-c color] ";
	print "[-right rd] [-left ld] [-level lvl] [-seed seed]\n";
        print " -help = this text\n";
        print "    -i = input file\n";
        print "    -o = output file\n";
	print "    -d = fig depth restriction\n";
	print "    -c = fig color restriction\n";
	print "-right = right indentation depth (default: 0.1443)\n";
	print " -left = left indentation depth (default: 0.0722)\n";
	print "-level = number of fractal iterations (default: 4)\n";
	print " -seed = random seed\n";
	exit;
    }
}

open(OUTDATA,">$Outfile");
open(INDATA,"<$Infile");

### SCAN FILE

while (<INDATA>) {
    if ((m/^2 1/) || (m/^2 3/)) {
	$hold = $_;
	($sub_type, $line_style, $thickness,
	 $pen_color, $fill_color, $depth,
	 $pen_style, $area_fill, $style_val,
	 $join_style, $cap_style, $radius,
	 $forward_arrow, $backward_arrow,
	 $npoints) =
	     m/^2 ([-]*\d+) ([-]*\d+) ([-]*\d+) ([-]*\d+) ([-]*\d+) ([-]*\d+) ([-]*\d+) ([-]*\d+) ([-]*\d+\.\d+) ([-]*\d+) ([-]*\d+) ([-]*\d+) ([-]*\d+) ([-]*\d+) ([-]*\d+)/;
	if ((($Depth eq -1) || ($Depth eq $depth)) &&
	    (($Color eq -2) || ($Color eq $pen_color))) {

### BEGIN MAIN LOOP
	    $j = 0;
	    @p = ();
	    while (<INDATA>) {
		if (!m/\t/) { last; }
		$i = 0;
		s/^[ \t\r\n]+//eg;
		s/[ \t\r\n]+/" "/eg;
		push @p,split(/ /);
		while ($p[$i]) {
		    if ($p[$i+3]) {
			coast($j++,$p[$i],$p[$i+1],$p[$i+2],$p[$i+3]);
		    }
		    else {
			@po;
			if ($p[$i+2]) {
			    $po[2] = $p[$i+2];
			    $po[1] = $p[$i+1];
			    $po[0] = $p[$i+0];
			}
			elsif ($p[$i+1]) {
			    $po[1] = $p[$i+1];
			    $po[0] = $p[$i+0];
			}
			elsif ($p[$i+0]) {
			    $po[0] = $p[$i+0];
			}
			@p = @po;
		    }
		    $i += 2;
		}
	    }
	    
	    $n = $j * 4**$LVL;
	    
	    foreach (0..$n-1) {
		my $r = int 2*sqrt(($x[$_+1]-$x[$_])**2+($y[$_+1]-$y[$_])**2);
		my $xc = int $x[$_];
		my $xr = int $x[$_] + $r;
		my $yc = int $y[$_];
		print OUTDATA "1 3 $line_style $thickness $pen_color ",
		"$fill_color $depth $pen_style ",
		"$area_fill $style_val 1 0.0000 ",
		"$xc $yc $r $r $xc $yc $xr $yc\n";
	    }
	    redo;
### END MAIN LOOP

	}
	else {
	    print OUTDATA $hold;
	}
    }
    else {
	print OUTDATA;
    }
}

sub coast { #j,x1,y1,x2,y2
    my $j = $_[0];

    $x[$j*4**$LVL] = $_[1];
    $y[$j*4**$LVL] = $_[2];
    $x[($j+1)*4**$LVL] = $_[3];
    $y[($j+1)*4**$LVL] = $_[4];

    f($j*4**$LVL,$LVL);

    sub f { #i, level
	my $i = $_[0];
	if ($_[1] ge 1) {
	    my $add = 4**($_[1]-1);
	    my $xd = 0;
	    my $yd = 0;
	    my $d = 0;
# lefthand normal
	    if (rand(2)-1 ge 0) {
		$yd = $IN*($x[$i+4*$add]-$x[$i]);
		$xd = -$IN*($y[$i+4*$add]-$y[$i]);
		$d = 0.25 + $IN*$IN;
	    }
	    else {
		$yd = -$OUT*($x[$i+4*$add]-$x[$i]);
		$xd = $OUT*($y[$i+4*$add]-$y[$i]);
		$d = 0.25 + $OUT*$OUT;
	    }
	    
	    $x[$i+1*$add] = (2*$x[$i] + 1*$x[$i+4*$add])/3;
	    $x[$i+2*$add] = ($x[$i] + $x[$i+4*$add])/2+$xd;
	    $x[$i+3*$add] = (1*$x[$i] + 2*$x[$i+4*$add])/3;
	    $y[$i+1*$add] = (2*$y[$i] + 1*$y[$i+4*$add])/3;
	    $y[$i+2*$add] = ($y[$i] + $y[$i+4*$add])/2+$yd;
	    $y[$i+3*$add] = (1*$y[$i] + 2*$y[$i+4*$add])/3;
	    
	    f($i+0*$add,$_[1]-1);
	    f($i+1*$add,$_[1]-1);
	    f($i+2*$add,$_[1]-1);
	    f($i+3*$add,$_[1]-1);
	}
    }
}
